(*
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2016)
License: LGPL (see file COPYING.LESSER)
*)
theory AC_Reduction_Pair_Processor
imports
  "../Framework/AC_DP_Framework"
  Usable_Rules
begin

lemma (in ce_af_redpair) i_trans_sound_dp_ac: 
  assumes
  fin_R: "finite R"
  and mode: "mode_cond mode m"
  and m: "m \<ge> n"
  and P: "(s,t) \<in> P"
  and steps: "(t \<cdot> \<sigma>, u \<cdot> \<tau>) \<in> (rstep RE)^*"
  and SN: "SN_on (relstep R E) {t \<cdot> \<sigma>}"
  and ur_closed: "ur_closed_af RE ur us \<pi>"
  and ur_P_closed: "ur_P_closed_af RE ur us \<pi> P"
  and orient: "ur \<subseteq> mode_left mode"
  and RE: "RE = R \<union> E"
  and sym_E: "symmetric_trs E"
  and AC_E: "AC_theory E F"
  shows  "(t \<cdot> itrans.i_trans_subst {} R RE E ur us (c,d,m) \<sigma>, 
    u \<cdot> itrans.i_trans_subst {} R RE E ur us (c,d,m) \<tau>) \<in> mode_NS mode"
proof -
  let ?cn = "(c,d,m)"
  note pre = fin_R RE sym_E AC_E
  interpret itrans "{}" R RE E F ur us ?cn by (rule itrans[OF pre])
  let ?s = "s \<cdot> \<sigma>"
  let ?t = "t \<cdot> \<sigma>"
  let ?u = "u \<cdot> \<tau>"
  let ?I = "i_trans"
  let ?Is = "i_trans_subst"
  let ?ssig = "s \<cdot> ?Is \<sigma>"
  let ?tsig = "t \<cdot> ?Is \<sigma>"
  let ?utau = "u \<cdot> ?Is \<tau>"
  let ?QR = "qrelac {} R E"
  have switch': "relstep R E = ?QR" unfolding qrelac_def  by auto
  have switch: "rstep RE = qrstep False {} R \<union> rstep E" unfolding RE by auto
  note SN' = SN[unfolded switch']
  note steps' = steps[unfolded switch]
  from SN_ac_preservation_steps[OF pre SN' steps'] 
  have SNu: "SN_on (qrelac {} R E) {u \<cdot> \<tau>}" .
  from i_transVI[OF pre mode m SN' steps' ur_closed orient]
  have one: "(?I ?t, ?I ?u) \<in> mode_NS mode" by simp
  from P ur_P_closed have "ur_closed_term_af RE ur us \<pi> t" by auto
  from i_transI[OF pre this SN'] have two: "(?tsig, ?I ?t) \<in> NS^*" by simp
  from i_transII[OF SNu, THEN conjunct1] ce_orient[OF m]
  have three: "(?I ?u, ?utau ) \<in> NS^*" by auto    
  from one two three have "(?tsig,?utau) \<in> NS^* O mode_NS mode O NS^*" by blast
  thus ?thesis by simp
qed

context
  fixes R E :: "('f,'v)trs" and F\<^sub>O\<^sub>C
  assumes fin_R: "finite R" 
  and E_lfun: "(Var x,r) \<notin> E"
  and root_E: "root_preserving_trs E"
  and symbol_E: "symbol_preserving_trs E"
  and sym_E: "symmetric_trs E"
  and AC_E: "AC_theory E F\<^sub>O\<^sub>C"
begin

lemma ac_redtriple_proc_main: 
  assumes ur: "ur \<subseteq> NS"
  and oP: "P \<subseteq> NST \<union> S"
  and D: "D \<subseteq> S" (* D = deleted pairs *)
  and redtriple: "ce_af_redtriple S NS NST \<pi>"
  and ur_closed: "ur_closed_af (R \<union> E) ur us \<pi>"
  and ur_P_closed: "ur_P_closed_af (R \<union> E) ur us \<pi> P"
  shows "finite_ac_dpp (D \<inter> P, P - D, {}, R, E)"
proof   
  def mode \<equiv> False
  fix s t \<sigma>
  assume chain: "min_iacchain (D \<inter> P, P - D, {}, R, E) s t \<sigma>"
  interpret ce_af_redtriple S NS NST \<pi> by fact
  let ?I = "itrans.i_trans_subst {} R (R \<union> E) E ur us (c,d,n)"
  def \<tau> \<equiv> "\<lambda> i. ?I (\<sigma> i)"
  let ?R = "rstep (R \<union> E)"
  have mode: "mode_cond mode n" unfolding mode_def mode_cond_def by auto
  have redpair: "ce_af_redpair S NS \<pi>" ..
  note chain = chain[unfolded min_iacchain_via_ichain ichain.simps minimal_cond_def]
  from chain have P: "\<And> i. (s i, t i) \<in> P" by auto
  from chain have steps: "\<And> i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> ?R^*" by auto
  from chain have SNt: "\<And> i. SN_on (relstep R E) {t i \<cdot> \<sigma> i}" by auto
  let ?s = "\<lambda> i. s i \<cdot> \<tau> i"
  let ?t = "\<lambda> i. t i \<cdot> \<tau> i"
  let ?next = "\<lambda> i. (s (Suc i), t (Suc i))"
  from ur have NS: "ur \<subseteq> mode_left mode" unfolding mode_def mode_left_def by auto
  from i_trans_sound_dp_ac[OF fin_R mode le_refl P steps SNt ur_closed ur_P_closed NS _ sym_E AC_E]
  have "\<And> i. (?t i, ?s (Suc i)) \<in> mode_NS mode" unfolding \<tau>_def by auto
  hence stepsNS: "\<And> i. (?t i, ?s (Suc i)) \<in> NS^*" unfolding mode_NS_def mode_def by auto
  from oP P have allP: "\<And> i. (s i, t i) \<in> NST \<union> S" by (auto simp: ichain.simps)
  from chain have inf: "INFM i. (s i, t i) \<in> D \<inter> P" by simp
  have piece1: "\<forall> i. (?t i, ?t (Suc i)) \<in> S \<or> (?t i, ?t (Suc i)) \<in> (NS \<union> NST)^* \<and> (?next i) \<notin> D"
  proof
    fix i
    show "(?t i, ?t (Suc i)) \<in> S \<or> (?t i, ?t (Suc i)) \<in> (NS \<union> NST)^* \<and> (?next i) \<notin> D" 
    proof (cases "?next i \<in> S")
      case True
      hence "(?s (Suc i), ?t (Suc i)) \<in> S" using subst_S by (auto simp: subst_closed_def)
      with stepsNS NS.trCompat show ?thesis by auto 
    next
      case False
      with stepsNS allP[of "Suc i"]
      have one: "?next i \<in> NST" and two: "?next i \<notin> S" by auto
      from one have "(?s (Suc i), ?t (Suc i)) \<in> NST" using subst_NST by (auto simp: subst_closed_def)
      with stepsNS[of i] have steps: "(?t i, ?t (Suc i)) \<in> NS^* O NST" by auto
      have "(?t i, ?t (Suc i)) \<in> (NS \<union> NST)^*"
        by (rule set_mp[OF _ steps], regexp)
      with two D show ?thesis by auto
    qed
  qed  
  hence infSeq: "\<forall> i. (?t i, ?t (Suc i)) \<in> (NS \<union> NST)^* \<union> S" by auto
  from SN have "SN_on S {?t 0}" by best
  from infSeq both.trCompat this have "\<exists> j. \<forall> i \<ge> j. (?t i, ?t (Suc i)) \<in> (NS \<union> NST)^* - S" by (rule non_strict_ending)  
  from this obtain j where one: "\<forall> i \<ge> j. (?t i, ?t (Suc i)) \<in> (NS \<union> NST)^* - S" ..
  with piece1 have ns: "\<forall> i \<ge> j. ?next i \<notin> D" by blast
  from inf[unfolded INFM_nat] obtain n where n: "n > j" and s: "(s n, t n) \<in> D" by auto
  from n obtain m where n: "n = Suc m" and m: "m \<ge> j" by (cases n, auto)
  from ns[THEN spec[of _ m]] s show False unfolding n using m by auto
qed

lemma ac_redtriple_proc: 
  assumes ur: "ur \<subseteq> NS"
  and PQ: "P \<union> Q \<subseteq> NST \<union> S"
  and D: "D \<subseteq> S"
  and redtriple: "ce_af_redtriple S NS NST \<pi>"
  and ur_closed: "ur_closed_af (R \<union> E) ur us \<pi>"
  and ur_P_closed: "ur_P_closed_af (R \<union> E) ur us \<pi> (P \<union> Q)"
  and R: "R = Rs \<union> Rw"
  and fin: "finite_ac_dpp (P - D, Q - D, Rs, Rw, E)"
  shows "finite_ac_dpp (P, Q, Rs, Rw, E)"
    by (rule finite_ac_dpp_split_top[OF fin], unfold R[symmetric], 
    rule ac_redtriple_proc_main[OF ur PQ D redtriple ur_closed ur_P_closed])
end


end